#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "curses.h"
#include "term.h"
#include "config.h"

#ifdef QDOS
#include <qdos.h>
char _prog_name[] = "tic";
char _version[]   = " v1.00 ";
char _copyright[]  = "(c)1991-1994  K.D.Walker";
void (*_consetup)() = consetup_title;           /* Fancy titled screen */
#endif /* QDOS */

#define MAGIC 0432    /* magic number for terminfo files */

#define NEWLINE 1
#define STRING  2
#define COMMA   3
#define HASH    4
#define EQUAL   5
#define AT      6
#define DOT     7
#define NAME    8

extern char *getenv();

static int optind = 1;

static int symbol;   /* the current symbol */
/*static int value;    /* symbol value (if NUMERIC) */
static char string[256];    /* symbol value (if STRING/NAME) */
static int ch;       /* current character */
static FILE *ifd;     /* current input file */
static int line = 0; /* the current input line */

/*
break the input stream up into a series of tokens which are
much more easily processed.
*/

static int
nextsym(void)
{
  static int new_line = TRUE;
  static int namestate = FALSE;
  char *s;

  while(TRUE)
    {
    switch (ch)
      {
    case ' ' :
    case '\t' : ch = fgetc(ifd); new_line = FALSE; break;
    case '\n' : new_line = TRUE; line++; ch = fgetc(ifd); break;
    case ',' : ch = fgetc (ifd); namestate = TRUE; return (COMMA);
    case '#' :
      if(new_line)
        {
        /* this is actually a comment line */
        for(; ch != '\n' && ch != EOF; ch = fgetc(ifd)) {}
        line++;
        break;
        }
      ch = fgetc (ifd);
      return (HASH);
    case '=' : ch = fgetc(ifd); return(EQUAL);
    case '@' : ch = fgetc(ifd); return(AT);
    case '.' : ch = fgetc(ifd); return(DOT);
    case EOF : return(EOF);
    default :
      if(new_line)
        {
        namestate = new_line = FALSE;
        line++;
        return(NEWLINE);
        }
      if(namestate)
        {
        for(s = string; isalnum(ch); ch = fgetc(ifd)) *s++ = ch;
        *s = '\0';
        namestate = FALSE;
        return(NAME);
        }
      for(s = string; ch!=',' && ch!=EOF; ch = fgetc(ifd))
        {
        if(ch == '\\') {*s++ = ch; ch = fgetc(ifd);}
        *s++ = ch;
        }
      *s = '\0';
      return(STRING);
      }
    }
}


/*
  This routine writes out the terminal description in a form
  which can be read by the routine setupterm().
*/

static int
write_term(TERMINAL *term, FILE *fd)
{
  char *sbptr, *ebptr;
  short *snptr, *enptr;
  char **ssptr, **esptr;
  short offset;
  short off;

  /* write out the header */

  fputc(MAGIC & 0xff, fd);    /* MAGIC number */
  fputc((MAGIC >> 8) & 0xff, fd);

  offset = strlen(term->_termdesc) +1;
  fputc(offset & 0xff, fd);
  fputc((offset >> 8) & 0xff, fd);

  offset = &(term->_bool_terminator) - &(term->_bool_starter) - 1;
  fputc(offset & 0xff, fd);
  fputc((offset >> 8) & 0xff, fd);

  offset = &(term->_numeric_terminator) - &(term->_numeric_starter) - 1;
  fputc(offset & 0xff, fd);
  fputc((offset >> 8) & 0xff, fd);

  offset = &(term->_string_terminator) - &(term->_string_starter) - 1;
  fputc(offset & 0xff, fd);
  fputc((offset >> 8) & 0xff, fd);

  for(offset = 0,
      ssptr = &term->_string_starter + 1,
      esptr = &term->_string_terminator;
      ssptr != esptr;
      ssptr++)
    {
    if(*ssptr && (*ssptr != (char *)-1) && (*ssptr != (char *)-2))
      offset += (strlen (*ssptr) + 1);
    }
  fputc(offset & 0xff, fd);
  fputc((offset >> 8) & 0xff, fd);

  /* write out the terminal name(s) */
  fputs(term->_termdesc, fd);
  fputc('\0', fd);

  /* write out the booleans */
  for(sbptr = &(term->_bool_starter) + 1,
      ebptr = &(term->_bool_terminator);
      sbptr != ebptr;
      sbptr++)
    {
    fputc(*sbptr, fd);
    }

  /* write out the integers */
  for(snptr = &(term->_numeric_starter) + 1,
      enptr = &(term->_numeric_terminator);
      snptr != enptr;
      snptr++)
    {
    fputc(*snptr & 0xff, fd);
    fputc((*snptr >> 8) & 0xff, fd);
    }

  /* write out the strings */
  for(off = 0,
      ssptr = &(term->_string_starter) + 1,
      esptr = &(term->_string_terminator);
      ssptr != esptr;
      ssptr++)
    {
    offset = (*ssptr && (*ssptr != (char *)-1) && (*ssptr != (char *)-2)) ? off : (int)*ssptr;
    fputc(offset & 0xff, fd);
    fputc((offset >> 8) & 0xff, fd);
    if(*ssptr && (*ssptr != (char *)-1) && (*ssptr != (char *)-2))
      off += (strlen (*ssptr) + 1);
    }

  /* write out the string table */
  for(ssptr = &term->_string_starter + 1,
      esptr = &term->_string_terminator;
      ssptr != esptr;
      ssptr++)
    {
    if(*ssptr && (*ssptr != (char *) -1) && (*ssptr != (char *) -2))
      {
      fputs(*ssptr,fd);
      fputc('\0', fd);
      }
    }

  return(OK);
}



/*
  define the numeric capability 'name' to the value 'num'.
*/

static void
addnumeric(TERMINAL *term, char *codename, int num)
  {
  char **  p;
  short *  v;

  for(p = numnames; *p && strcmp(*p, codename); p++) {}

  if(*p)
    {
    v = &term->_numeric_starter + (p-numnames) + 1;
    if(*v >= 0)
      fprintf(stderr,"numeric capability '%s' multiply defined\n", codename);
    *v = num;
    }
  else
    fprintf(stderr, "numeric capabiliy '%s' not found\n", codename);
}



/*
  define the string capability 'name' to contain 'string'.
*/

static void
addstring(TERMINAL *term, char *codename, char *str)
{
  char **p;
  char *r;
  char *s;

  for(p = strnames; *p && strcmp(*p, codename); p++) {}

  if(*p)
    {
    p = &term->_string_starter + (p-strnames) + 1;

    /* must process the string */
    for(s = r = str; *s; s++, r++)
      {
      switch(*s)
        {
      case '^' : s++; *r = (*s & 037); break;
      case '\\' :
        s++;
        switch(*s)
          {
        case 'b' : *r = '\b'; break;
        case 'E' :
        case 'e' : *r = '\033'; break;
        case 'f' : *r = '\f'; break;
        case 'n' : *r = '\n'; break;
        case 'r' : *r = '\r'; break;
        case 's' : *r = ' '; break;
        case 't' : *r = '\t'; break;
        case '^' :
        case '\\':
        case ',' :
        case ':' : *r = *s; break;

        case '0' :
        case '1' :
        case '2' :
        case '3' :
        case '4' :
        case '5' :
        case '6' :
        case '7' :
          *r = *s - '0';
          switch (*(s+1))
            {
          case '0' :
          case '1' :
          case '2' :
          case '3' :
          case '4' :
          case '5' :
          case '6' :
          case '7' :
            *r = *r*8 + *(++s) - '0';
            switch (*(s+1))
              {
            case '0' :
            case '1' :
            case '2' :
            case '3' :
            case '4' :
            case '5' :
            case '6' :
            case '7' :
              *r = *r*8 + *(++s) - '0';
              }
            }
          }
        if(*r==0) *r = 0x80;
        break;
      default :
        *r = *s;
        }
      }
    *r = '\0';

    if(*p && *p != (char *)-1 && *p != (char *)-2)
      {
      free(*p);
      fprintf(stderr, "string capability '%s' multiply defined\n", codename);
      }
    if((*p=malloc(strlen(str)+1))!=0) strcpy(*p, str);
    }
  else
    fprintf (stderr, "string capability '%s' not found\n", codename);
}


/*
  define the boolean capability 'name'.
*/


static void
addbool(TERMINAL *term, char *codename)
{
  char **p;
  char *v;

  for(p = boolnames; *p && strcmp (*p, codename); p++) {}

  if(*p)
    {
    v = &term->_bool_starter + (p-boolnames) + 1;
    if(*v != 0)
      fprintf(stderr,"boolean capability '%s' multiply defined\n",codename);
    *v = 1;
    }
  else
    fprintf(stderr, "boolean capabiliy '%s' not found\n", codename);
}


/*
  This routine searches for a capability 'name' and then
  undefines this capability.
*/

static void
deletecap(TERMINAL *term, char *codename)
  {
  char **p;

  for(p = boolnames; *p && strcmp (*p, codename); p++) {}

  if(*p)
    {
    *(&term->_bool_starter + (p-boolnames) + 1) = 2;
    return;
    }

  for(p = numnames; *p && strcmp (*p, codename); p++) {}

  if(*p)
    {
    *(&term->_numeric_starter + (p-numnames) + 1) = -2;
    return;
    }

  for(p = strnames; *p && strcmp (*p, codename); p++) {}

  if(*p)
    {
    *(&term->_string_starter + (p-strnames) + 1) = (char *)-2;
    return;
    }

  fprintf(stderr,"unable to delete unknown capability '%s'\n", codename);
}



static TERMINAL *
create_terminal(char *names)
{
  TERMINAL *term;
  char *bstart, *bend;
  short *nstart, *numend;
  char **sstart, **strend;


  if(!names) return((TERMINAL *)0);

  if((term=(TERMINAL*)malloc(sizeof(TERMINAL)))==0) return((TERMINAL*)0);

  term->_termdesc = malloc(strlen(names)+1);
  strcpy(term->_termdesc, names);

  for(bstart = &term->_bool_starter + 1,
      bend = &term->_bool_terminator;
      bstart != bend;
      bstart++)
    {
    *bstart = 0;
    }

  for(nstart = &term->_numeric_starter + 1,
      numend = &term->_numeric_terminator;
      nstart != numend;
      nstart++)
    {
    *nstart = -1;
    }

  for(sstart = &term->_string_starter + 1,
      strend = &term->_string_terminator;
      sstart != strend;
      sstart++)
    {
    *sstart = (char *) -1;
    }
  return(term);
}


static void
free_definition(TERMINAL *term)
  {
  char **sstart;
  char **strend;

  if(!term) return;

  for(sstart = &term->_string_starter + 1,
      strend = &term->_string_terminator;
      sstart != strend;
      sstart++)
    {
    if(*sstart && *sstart != (char *)-1 && *sstart != (char *)-2)
      free(*sstart);
    }
  free(term->_termdesc);
  free(term);
}


/*
  Read in the defintion for the terminal 'name' and copy across all
  those intries which have not yet been set in 'term'
*/

static void
use_terminal(TERMINAL *term, char *name)
{
  char *bstart, *bend, *bstart2;
  short *nstart, *numend, *nstart2;
  char **sstart, **strend, **sstart2;
  int errret;

  if(setupterm(name, 1, &errret) == ERR)
    {
    fprintf(stderr,"terminal '%s' referenced in use statement not found\n", name);
    exit(1);
    }

  for(bstart = &term->_bool_starter + 1,
      bend = &term->_bool_terminator,
      bstart2 = &cur_term->_bool_starter + 1;
      bstart != bend;
      bstart++, bstart2++)
    {
    if(*bstart == 0) *bstart = *bstart2;
    }

  for(nstart = &term->_numeric_starter + 1,
      numend = &term->_numeric_terminator,
      nstart2 = &cur_term->_numeric_starter + 1;
      nstart != numend;
      nstart++, nstart2++)
    {
    if(*nstart == -1) *nstart = *nstart2;
    }

  for(sstart = &term->_string_starter + 1,
      strend = &term->_string_terminator,
      sstart2 = &cur_term->_string_starter + 1;
      sstart != strend;
      sstart++, sstart2++)
    {
    if(*sstart == (char *)-1 || *sstart == (char *)-2)
      {
      if(!*sstart2)
        *sstart = (char * )-1;
      else if(*sstart2 == (char *)-1 || *sstart2 == (char *)-2)
        *sstart = *sstart2;
      else
        {
        if((*sstart=malloc(strlen(*sstart2)+1))!=0)
          strcpy(*sstart, *sstart2);
        }
      }
    }

  del_curterm(cur_term);
  cur_term = (TERMINAL *)NULL;
}


/*
  Parse the terminal name line and write out a definition
  for each name specified.  The terminal names are separated
  by a vertical bar.  If there is only one name then it is
  used as the name of the terminal.  If there is more than
  one name then the last names is actually a verbose
  description of the terminal and not really a terminal name.
*/

static void
write_definition(TERMINAL *term, char *names)
{
  FILE *fd;
  char *terminfo, *name;
  char termnam[256], aliasname[256];

  if(!term || !names || !*names) return;

  /* remove the longname */
  for (name = names+strlen(names); *(--name)!='|';) {}
  *name = '\0';
  name = strtok(names,"|");

  if((terminfo = getenv ("TERMINFO"))==0)
    terminfo = TERMDIR;

  strcpy(termnam, terminfo);
  strcat(termnam, DIR_SEPERATOR);
  strcat(termnam, name);

  if((fd = fopen(termnam, "w"))==0)
    {
    fprintf(stderr, "cannot open ouput file '%s'\n", termnam);
    return;
    }
  write_term(term, fd);
  fclose(fd);

  /* creating the terminal aliases */
  while(name = strtok((char *)0, "|"))
    {
    strcpy(aliasname, terminfo);
    strcat(aliasname, DIR_SEPERATOR);
    strcpy(aliasname, name);
#ifndef QDOS
    link(termnam,aliasname);
#else
  if((fd = fopen(aliasname, "w"))==0)
    {
    fprintf(stderr, "cannot open ouput file '%s'\n", aliasname);
    }
  write_term(term, fd);
  fclose(fd);
#endif
    }
}


/*
  This routine checks that the current symbol is the symbol
  that we are expecting.  If it is proceed to the next
  symbol.  If it is not skip until a comma is found.
*/

static void
expect(int sym)
{
  if(sym == symbol)
    {
    symbol = nextsym();
    return;
    }

  /* OK, we have an error */
  fprintf(stderr, "syntax error on line %d\n", line);
  fprintf(stderr, "%d expected, got %d\n", sym, symbol);
  fprintf(stderr, "symbol is %s\n", string);
  exit(1);
}



/*
  This routine parses a complete terminal definition
*/

static void
parse(void)
{
  TERMINAL *term;
  char termnames[256];
  char name[256];
  char str[256];

  while(symbol != EOF)
    {
    expect(NEWLINE);

    /* terminal name(s) */
    strcpy(termnames, string);
    expect(STRING);
    expect(COMMA);
    term = create_terminal(termnames);

    while(symbol != EOF && symbol != NEWLINE)
      {
      switch(symbol)
        {
      case NAME :
        /* capability name */
        strcpy(name, string);
        expect(NAME);
        switch(symbol)
          {
        case HASH :
          /* numeric capability */
          expect(HASH);
          strcpy(str, string);
          expect(STRING);
          expect(COMMA);
          addnumeric(term, name, (int)atol(str));
          break;
        case EQUAL :
          /* string capability */
          expect(EQUAL);
          strcpy(str, string);
          expect(STRING);
          expect(COMMA);
          if(strcmp(name, "use")) addstring(term, name, str);
          else use_terminal(term, str);
          break;
        case COMMA :
          /* boolean capability */
          expect(COMMA);
          addbool(term, name);
          break;
        case AT :
          /* capability deleted */
          expect(AT);
          expect(COMMA);
          deletecap(term, name);
          break;
          }
        break;
      case DOT :
        /* capability commented out */
        while(symbol != COMMA && symbol != EOF) expect(symbol);
        expect(COMMA);
        break;
      case STRING:
        fprintf(stderr, "unexpected string '%s' at line %d\n", string, line);
        symbol = nextsym();
        break;
      default :
        fprintf(stderr, "unexpected symbol %d at line %d\n", symbol, line);
        symbol = nextsym();
        }
      }
    /* definition must now be completed */
    write_definition(term, termnames);
    free_definition(term);
    term = (TERMINAL *)NULL;
    }
}


int
main(int argc, char *argv[])
  {
  if(optind >= argc)
    {
    fprintf(stderr, "%s: no source file specified\n", *argv);
    exit(1);
    }

  for(; optind < argc; optind++)
    {
    if((ifd = fopen(argv[optind], "r"))==0)
      {
      fprintf(stderr, "%s: cannot open '%s'\n", *argv, argv[optind]);
      exit(2);
      }
    ch = fgetc(ifd);
    symbol = nextsym();
    parse();
    fclose(ifd);
    }

  return(0);
}
